@@ -15,6 +15,30 @@
 # 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 #
 
+## @file
+# The python module to implement things needed by gstlal_inspiral
+#
+# ### Review Status
+#
+# STATUS: reviewed with actions
+#
+# | Names                                          | Hash                                        | Date       |
+# | -------------------------------------------    | ------------------------------------------- | ---------- |
+# | Kipp Cannon, Chad Hanna, Jolien Creighton, Florent Robinet, B. Sathyaprakash, Duncan Meacher, T.G.G. Li | b8fef70a6bafa52e3e120a495ad0db22007caa20 | 2014-12-03 |
+#
+# #### Action items
+# - Document examples of how to get SNR history, etc., to a web browser in an offline search
+# - Long term goal: Using template duration (rather than chirp mass) should load balance the pipeline and improve statistics
+# - L651: One thing to sort out is the signal probability while computing coincs
+# - L640-L647: Get rid of obsolete comments
+# - L667: Make sure timeslide events are not sent to GRACEDB
+# - Lxxx: Can normalisation of the tail of the distribution pre-computed using fake data?
+# - L681: fmin should not be hard-coded to 10 Hz. horizon_distance will be horribly wrong if psd is constructed, e.g. using some high-pass filter. For example, change the default to 40 Hz.
+# - L817: If gracedb upload failed then it should be possible to identify the failure, the specifics of the trigger that encountered failure and a way of submitting the trigger again to gracedb is important. Think about how to clean-up failures.
+# - Mimick gracedb upload failures and see if the code crashes
+
+
+## @package inspiral
 
 #
 # =============================================================================
@@ -44,6 +68,7 @@ import threading
 import time
 import httplib
 import tempfile
+import shutil
 
 # The following snippet is taken from http://gstreamer.freedesktop.org/wiki/FAQ#Mypygstprogramismysteriouslycoredumping.2Chowtofixthis.3F
 import pygtk
@@ -74,8 +99,9 @@ from glue.ligolw.utils import process as ligolw_process
 from glue.ligolw.utils import search_summary as ligolw_search_summary
 from glue.ligolw.utils import segments as ligolw_segments
 from glue.ligolw.utils import time_slide as ligolw_time_slide
-from pylal import datatypes as laltypes
+import lal
 from pylal import rate
+from pylal.datatypes import LALUnit
 from pylal.datatypes import LIGOTimeGPS
 from pylal.datatypes import REAL8FrequencySeries
 from pylal.xlal.datatypes.snglinspiraltable import from_buffer as sngl_inspirals_from_buffer
@@ -261,6 +287,15 @@ def parse_iirbank_files(iir_banks, verbose, snr_threshold = 4.0):
 	return banks
 
 
+def subdir_from_T050017_filename(fname):
+	path = str(fname.split("-")[2])[:5]
+	try:
+		os.mkdir(path)
+	except OSError:
+		pass
+	return path
+
+
 #
 # =============================================================================
 #
@@ -438,16 +473,16 @@ class CoincsDocument(object):
 
 	@property
 	def search_summary_outseg(self):
-		return self.search_summary.get_out()
+		return self.search_summary.out_segment
 
 
 	def add_to_search_summary_outseg(self, seg):
 		out_segs = segments.segmentlist([self.search_summary_outseg])
-		if out_segs == [segments.segment(None, None)]:
+		if out_segs == [None]:
 			# out segment not yet initialized
 			del out_segs[:]
 		out_segs |= segments.segmentlist([seg])
-		self.search_summary.set_out(out_segs.extent())
+		self.search_summary.out_segment = out_segs.extent()
 
 
 	def get_next_sngl_id(self):
@@ -457,7 +492,7 @@ class CoincsDocument(object):
 	def T050017_filename(self, description, extension):
 		start, end = self.search_summary_outseg
 		start, end = int(math.floor(start)), int(math.ceil(end))
-		return "%s-%s-%d-%d.%s" % ("".join(sorted(self.process.get_ifos())), description, start, end - start, extension)
+		return "%s-%s-%d-%d.%s" % ("".join(sorted(self.process.instruments)), description, start, end - start, extension)
 
 
 	def write_output_file(self, verbose = False):
@@ -469,7 +504,7 @@ class CoincsDocument(object):
 			# record the final state of the search_summary and
 			# process rows in the database
 			cursor = self.connection.cursor()
-			if seg != segments.segment(None, None):
+			if seg is not None:
 				cursor.execute("UPDATE search_summary SET out_start_time = ?, out_start_time_ns = ?, out_end_time = ?, out_end_time_ns = ? WHERE process_id == ?", (seg[0].seconds, seg[0].nanoseconds, seg[1].seconds, seg[1].nanoseconds, self.search_summary.process_id))
 			cursor.execute("UPDATE search_summary SET nevents = (SELECT count(*) FROM sngl_inspiral) WHERE process_id == ?", (self.search_summary.process_id,))
 			cursor.execute("UPDATE process SET end_time = ? WHERE process_id == ?", (self.process.end_time, self.process.process_id))
@@ -486,7 +521,7 @@ class CoincsDocument(object):
 
 
 class Data(object):
-	def __init__(self, filename, process_params, pipeline, instruments, seg, coincidence_threshold, coinc_params_distributions, ranking_data, marginalized_likelihood_file = None, likelihood_file = None, injection_filename = None, time_slide_file = None, comment = None, tmp_path = None, likelihood_snapshot_interval = None, thinca_interval = 50.0, sngls_snr_threshold = None, gracedb_far_threshold = None, gracedb_group = "Test", gracedb_search = "LowMass", gracedb_pipeline = "gstlal", replace_file = True, verbose = False):
+	def __init__(self, filename, process_params, pipeline, instruments, seg, coincidence_threshold, coinc_params_distributions, ranking_data, marginalized_likelihood_file = None, likelihood_files_namedtuple = None, injection_filename = None, time_slide_file = None, comment = None, tmp_path = None, likelihood_snapshot_interval = None, thinca_interval = 50.0, sngls_snr_threshold = None, gracedb_far_threshold = None, gracedb_group = "Test", gracedb_search = "LowMass", gracedb_pipeline = "gstlal", gracedb_service_url = "https://gracedb.ligo.org/api/", replace_file = True, verbose = False):
 		#
 		# initialize
 		#
@@ -496,7 +531,7 @@ class Data(object):
 		self.verbose = verbose
 		# None to disable likelihood ratio assignment, otherwise a filename
 		self.marginalized_likelihood_file = marginalized_likelihood_file
-		self.likelihood_file = likelihood_file
+		self.likelihood_files_namedtuple = likelihood_files_namedtuple
 		# None to disable periodic snapshots, otherwise seconds
 		# set to 1.0 to disable background data decay
 		self.likelihood_snapshot_interval = likelihood_snapshot_interval
@@ -506,6 +541,7 @@ class Data(object):
 		self.gracedb_group = gracedb_group
 		self.gracedb_search = gracedb_search
 		self.gracedb_pipeline = gracedb_pipeline
+		self.gracedb_service_url = gracedb_service_url
 
 		#
 		# setup bottle routes
@@ -599,6 +635,19 @@ class Data(object):
 				# somehow, no?
 				self.pipeline.get_bus().post(message_new_checkpoint(self.pipeline, timestamp = buf_timestamp.ns()))
 
+				# If a reference likelihood file is given, load
+				# data coinc_params_distributions data before
+				# smoothing distributions
+				# FIXME There is currently no guarantee that
+				# the reference_likelihood_file on disk will
+				# have updated since the last snapshot, but for
+				# our purpose it should not have that large of
+				# an effect. The data loaded should never be
+				# older than the snapshot before last
+				if self.likelihood_files_namedtuple.reference_likelihood_file is not None:
+					self.coinc_params_distributions = far.parse_likelihood_control_doc(ligolw_utils.load_filename(self.likelihood_files_namedtuple.reference_likelihood_file, verbose = self.verbose, contenthandler = far.ThincaCoincParamsDistributions.LIGOLWContentHandler))[0]
+					self.coinc_params_distributions.finish(verbose = self.verbose)
+
 				if self.marginalized_likelihood_file is not None:
 					# FIXME:  must set horizon
 					# distances in coinc params object
@@ -613,22 +662,12 @@ class Data(object):
 					# ratio distributions that have
 					# been updated asynchronously and
 					# initialize a FAP/FAR assignment
-					# machine from it.  NOTE:  to keep
-					# overhead low we do not .finish()
-					# the ranking data here;  the
-					# external process generating this
-					# input file must do that for us.
-					coinc_params_distributions, ranking_data, seglists = far.parse_likelihood_control_doc(ligolw_utils.load_filename(self.marginalized_likelihood_file, verbose = self.verbose, contenthandler = far.ThincaCoincParamsDistributions.LIGOLWContentHandler))
-					if coinc_params_distributions is None:
-						raise ValueError("\"%s\" does not contain coinc parameter PDFs" % self.marginalized_likelihood_file)
+					# machine from it.
+					ranking_data, seglists = far.parse_likelihood_control_doc(ligolw_utils.load_filename(self.marginalized_likelihood_file, verbose = self.verbose, contenthandler = far.ThincaCoincParamsDistributions.LIGOLWContentHandler))[1:]
 					if ranking_data is None:
 						raise ValueError("\"%s\" does not contain ranking statistic PDFs" % self.marginalized_likelihood_file)
-					# we're using the class attribute
-					# elsewhere so make sure these two
-					# match
-					assert ranking_data.ln_likelihood_ratio_threshold == far.RankingData.ln_likelihood_ratio_threshold
 					ranking_data.finish(verbose = self.verbose)
-					self.fapfar = far.FAPFAR(ranking_data, coinc_params_distributions.count_above_threshold, threshold = far.RankingData.ln_likelihood_ratio_threshold, livetime = far.get_live_time(seglists))
+					self.fapfar = far.FAPFAR(ranking_data, livetime = far.get_live_time(seglists))
 
 			# run stream thinca.  update the parameter
 			# distribution data from sngls that weren't used in
@@ -678,7 +717,8 @@ class Data(object):
 
 	def record_horizon_distance(self, instrument, timestamp, psd, m1, m2, snr_threshold = 8.0):
 		with self.lock:
-			horizon_distance = reference_psd.horizon_distance(psd, m1 = m1, m2 = m2, snr = snr_threshold, f_min = 10.0, f_max = 0.85 * (psd.f0 + (len(psd.data) - 1) * psd.deltaF))
+			# FIXME get frequency bounds from templates
+			horizon_distance = reference_psd.horizon_distance(psd, m1 = m1, m2 = m2, snr = snr_threshold, f_min = 40.0, f_max = 0.85 * (psd.f0 + (len(psd.data) - 1) * psd.deltaF))
 			assert not (math.isnan(horizon_distance) or math.isinf(horizon_distance))
 			# NOTE:  timestamp is cast to float.  should be
 			# safe, whitener should be reporting PSDs with
@@ -727,6 +767,26 @@ class Data(object):
 				if (coinc_event.likelihood >= far.RankingData.ln_likelihood_ratio_threshold or self.marginalized_likelihood_file is None) and not any(offset_vector.values()):
 					self.coinc_params_distributions.add_zero_lag(self.coinc_params_distributions.coinc_params(self.stream_thinca.last_coincs.sngl_inspirals(coinc_event_id), offset_vector))
 
+		# Cluster last coincs before recording number of zero
+		# lag events or sending alerts to gracedb
+		# FIXME Do proper clustering that saves states between
+		# thinca intervals and uses an independent clustering
+		# window. This can also go wrong if there are multiple
+		# events with an identical likelihood.  It will just
+		# choose the event with the highest event id
+		if self.stream_thinca.last_coincs:
+			self.stream_thinca.last_coincs.coinc_event_index = dict([max(self.stream_thinca.last_coincs.coinc_event_index.iteritems(), key = lambda (coinc_event_id, coinc_event): coinc_event.likelihood)])
+
+		# Add events to the observed likelihood histogram post "clustering"
+		# FIXME proper clustering is really needed (see above)
+		if self.stream_thinca.last_coincs:
+			for coinc_event_id, coinc_event in self.stream_thinca.last_coincs.coinc_event_index.items():
+				offset_vector = self.stream_thinca.last_coincs.offset_vector(coinc_event.time_slide_id)
+				#IFOS come from coinc_inspiral not coinc_event
+				ifos = self.stream_thinca.last_coincs.coinc_inspiral_index[coinc_event_id].ifos
+				if (coinc_event.likelihood is not None and coinc_event.likelihood >= far.RankingData.ln_likelihood_ratio_threshold) and not any(offset_vector.values()):
+					self.ranking_data.zero_lag_likelihood_rates[frozenset(lsctables.instrument_set_from_ifos(ifos))][coinc_event.likelihood,] += 1
+
 		# do GraceDB alerts
 		if self.gracedb_far_threshold is not None:
 			self.__do_gracedb_alerts()
@@ -737,7 +797,7 @@ class Data(object):
 
 	def __do_gracedb_alerts(self):
 		if self.stream_thinca.last_coincs:
-			gracedb_client = gracedb.Client()
+			gracedb_client = gracedb.Client(url=self.gracedb_service_url)
 			gracedb_ids = []
 			psdmessage = None
 			coinc_inspiral_index = self.stream_thinca.last_coincs.coinc_inspiral_index
@@ -765,12 +825,17 @@ class Data(object):
 					psddict = {}
 					for instrument in self.seglistdicts["triggersegments"]:
 						elem = self.pipeline.get_by_name("lal_whiten_%s" % instrument)
+						# FIXME:  remove
+						# LIGOTimeGPS type cast
+						# when we port to swig
+						# version of
+						# REAL8FrequencySeries
 						psddict[instrument] = REAL8FrequencySeries(
 							name = "PSD",
-							epoch = LIGOTimeGPS(0, 0),	# FIXME
+							epoch = LIGOTimeGPS(lal.UTCToGPS(time.gmtime()), 0),
 							f0 = 0.0,
 							deltaF = elem.get_property("delta-f"),
-							sampleUnits = laltypes.LALUnit("s strain^2"),	# FIXME:  don't hard-code this
+							sampleUnits = LALUnit("s strain^2"),	# FIXME:  don't hard-code this
 							data = numpy.array(elem.get_property("mean-psd"))
 						)
 					psdmessage = StringIO.StringIO()
@@ -935,7 +1000,7 @@ class Data(object):
 		except:
 			yield "error\n"
 
-	def __write_output_file(self, filename = None, likelihood_file = None, verbose = False):
+	def __write_output_file(self, filename = None, verbose = False):
 		self.__flush()
 
 		# FIXME:  should this be done in .flush() somehow?
@@ -946,6 +1011,7 @@ class Data(object):
 			self.coincs_document.filename = filename
 		self.coincs_document.write_output_file(verbose = verbose)
 
+	def __write_likelihood_file(self, filename, description, snapshot = False, verbose = False):
 		# write the parameter PDF file.  NOTE;  this file contains
 		# raw bin counts, and might or might not contain smoothed,
 		# normalized, PDF arrays but if it does they will not
@@ -956,27 +1022,37 @@ class Data(object):
 		# thinca.  we want to know precisely when the arrays get
 		# updated so we can have a hope of computing the likelihood
 		# ratio PDFs correctly.
-		if likelihood_file is None:
-			likelihood_file = os.path.split(self.coincs_document.filename)
-			try: # preserve LIGO-T010150-00 if possible 
-				ifo, desc, start, dur = ".".join(likelihood_file[1].split('.')[:-1]).split('-')
-				likelihood_file = os.path.join(likelihood_file[0], '%s-%s_SNR_CHI-%s-%s.xml.gz' % (ifo, desc, start, dur))
-			except ValueError:
-				likelihood_file = os.path.join(likelihood_file[0], '%s_SNR_CHI.xml.gz' % likelihood_file[1].split('.')[0])
-		ligolw_utils.write_filename(self.__get_likelihood_file(), likelihood_file, gz = (likelihood_file or "stdout").endswith(".gz"), verbose = verbose, trap_signals = None)
 
-		# can't be used anymore
-		del self.coincs_document
+		path, filename = os.path.split(filename)
+		tmp_likelihood_file = os.path.join(path, 'tmp_%s' % filename)
+		ligolw_utils.write_filename(self.__get_likelihood_file(), tmp_likelihood_file, gz = (filename or "stdout").endswith(".gz"), verbose = verbose, trap_signals = None)
+		shutil.move(tmp_likelihood_file, os.path.join(path,filename))
+		# Snapshots get their own custom file and path
+		if snapshot:
+			fname = self.coincs_document.T050017_filename(description + '_DISTSTATS', 'xml.gz')
+			shutil.copy(os.path.join(path,filename), os.path.join(subdir_from_T050017_filename(fname), fname))
 
-	def write_output_file(self, filename = None, likelihood_file = None, verbose = False):
+	def write_output_file(self, filename = None, description = "", verbose = False):
 		with self.lock:
-			self.__write_output_file(filename = filename, likelihood_file = likelihood_file, verbose = verbose)
+			self.__write_output_file(filename = filename, verbose = verbose)
+			if self.likelihood_files_namedtuple.likelihood_file: 
+				self.__write_likelihood_file(self.likelihood_files_namedtuple.likelihood_file, description, verbose = verbose)
+
+			# can't be used anymore
+			del self.coincs_document
 
 	def snapshot_output_file(self, description, extension, verbose = False):
 		with self.lock:
 			coincs_document = self.coincs_document.get_another()
 			# We require the likelihood file to have the same name
 			# as the input to this program to accumulate statistics
-			# as we go 
-			self.__write_output_file(filename = self.coincs_document.T050017_filename(description, extension), likelihood_file = self.likelihood_file, verbose = verbose)
+			# as we go
+			fname = self.coincs_document.T050017_filename(description, extension)
+			fname = os.path.join(subdir_from_T050017_filename(fname), fname)
+			self.__write_output_file(filename = fname, verbose = verbose)
+			if self.likelihood_files_namedtuple.likelihood_file:
+				self.__write_likelihood_file(self.likelihood_files_namedtuple.likelihood_file, description, snapshot = True, verbose = verbose)
+
+			# can't be used anymore
+			del self.coincs_document
 			self.coincs_document = coincs_document
