@@ -65,12 +65,13 @@
 # =============================================================================
 #
 
-
+from copy import deepcopy
 import math
-import sys
 import numpy
-import warnings
+import os
 import StringIO
+import sys
+import warnings
 
 
 # The following snippet is taken from http://gstreamer.freedesktop.org/wiki/FAQ#Mypygstprogramismysteriouslycoredumping.2Chowtofixthis.3F
@@ -95,6 +96,7 @@ from gstlal import multirate_datasource
 from gstlal import pipeio
 from gstlal import pipeparts
 from gstlal import simplehandler
+import lal
 from pylal import series as lalseries
 from pylal.datatypes import LIGOTimeGPS
 
@@ -205,7 +207,7 @@ class Handler(simplehandler.Handler):
 	dumps of segment information, trigger files and background
 	distribution statistics.
 	"""
-	def __init__(self, mainloop, pipeline, dataclass, instruments, tag = "", seglistdict = None, verbose = False):
+	def __init__(self, mainloop, pipeline, dataclass, instruments, tag = "", seglistdict = None, segment_history_duration = 2592000., verbose = False):
 		"""!
 		@param mainloop The main application's event loop
 		@param pipeline The gstreamer pipeline that is being controlled by this handler
@@ -218,6 +220,7 @@ class Handler(simplehandler.Handler):
 		self.dataclass = dataclass
 
 		self.tag = tag
+		self.segment_history_duration = segment_history_duration
 		self.verbose = verbose
 
 		# setup segment list collection from gates
@@ -234,11 +237,7 @@ class Handler(simplehandler.Handler):
 		# record, or maybe not because it could result in a lot of
 		# duplication of on-disk data.  who knows.  think about it.
 		gate_suffix = {
-			# FIXME:  datasegments temporarily disabled until
-			# we decide how to deal with livetime in dag (which
-			# is currently adding its own datasegments lists to
-			# the final database)
-			#"datasegments": "frame_segments_gate",
+			"framesegments": "frame_segments_gate",
 			"statevectorsegments": "state_vector_gate",
 			"whitehtsegments": "ht_gate"
 		}
@@ -256,6 +255,9 @@ class Handler(simplehandler.Handler):
 		# FIXME:  don't do this, get rid of the Data class
 		dataclass.seglistdicts = self.seglistdicts
 
+		# create a deep copy to keep track of cumulative segments
+		self.cumulative_seglistdicts = deepcopy(self.seglistdicts)
+
 		# state of segments being collected
 		self.current_segment_start = {}
 
@@ -292,6 +294,7 @@ class Handler(simplehandler.Handler):
 
 		# segment lists
 		bottle.route("/segments.xml")(self.web_get_segments_xml)
+		bottle.route("/cumulative_segments.xml")(self.web_get_cumulative_segments_xml)
 
 	def do_on_message(self, bus, message):
 		"""!
@@ -324,6 +327,12 @@ class Handler(simplehandler.Handler):
 			if message.structure.get_name() == "CHECKPOINT":
 				# FIXME make a custom parser for CHECKPOINT messages?
 				timestamp = message.structure["timestamp"]
+				# FIXME:  the function that makes these
+				# messages uses a default value of None,
+				# and in principle could be called with
+				# other non-timestamp-like things, so this
+				# code should check that it has actually
+				# gotten a valid timestamp
 				self.checkpoint(timestamp)
 				return True
 		elif message.type == gst.MESSAGE_EOS:
@@ -339,6 +348,11 @@ class Handler(simplehandler.Handler):
 		return False
 
 	def _close_segments(self, timestamp):
+		"""
+		@timestamp must be a GPS time that is guaranteed to precede
+		any possible future state transitions in all segment lists
+		being tracked.
+		"""
 		# close out existing segments.  the code in the loop
 		# modifies the iteration target, so iterate over a copy
 		for (segtype, instrument), start_time in list(self.current_segment_start.items()):
@@ -368,6 +382,13 @@ class Handler(simplehandler.Handler):
 
 		@param timestamp the gstreamer timestamp in nanoseconds of the current buffer in order to close off open segment intervals before writing to disk
 		"""
+		# FIXME:  the timestamp is used to close off open segments
+		# and so should *not* be the timestamp of the current
+		# buffer, necessarily, but rather a GPS time guaranteed to
+		# precede any future state transitions of any segment list.
+		# especially if the pipeline is ever run in an "advanced
+		# warning" configuration using the GPS time of a trigger
+		# buffer would be an especially bad choice.
 		self.flush_segments_to_disk(timestamp)
 		try:
 			self.dataclass.snapshot_output_file("%s_LLOID" % self.tag, "xml.gz", verbose = self.verbose)
@@ -383,11 +404,18 @@ class Handler(simplehandler.Handler):
 		"""
 		with self.dataclass.lock:
 			try:
-				# close out existing segments.
+				# close out existing and update cumulative segments.
 				self._close_segments(timestamp)
+				self.update_cumulative_segments()
 				ext = segments.segmentlist(seglistdict.extent_all() for seglistdict in self.seglistdicts.values()).extent()
 				instruments = set(instrument for seglistdict in self.seglistdicts.values() for instrument in seglistdict)
-				fname = "%s-%s_SEGMENTS-%d-%d.xml.gz" % ("".join(sorted(instruments)), self.tag, int(math.floor(ext[0])), int(math.ceil(ext[1])) - int(math.floor(ext[0])))
+				#FIXME integrate with the Data class snapshotting directories
+				path = str(int(math.floor(ext[0])))[:5]
+				try:
+					os.mkdir(path)
+				except OSError:
+					pass
+				fname = "%s/%s-%s_SEGMENTS-%d-%d.xml.gz" % (path, "".join(sorted(instruments)), self.tag, int(math.floor(ext[0])), int(math.ceil(ext[1])) - int(math.floor(ext[0])))
 				ligolw_utils.write_filename(self.gen_segments_xmldoc(), fname, gz = fname.endswith('.gz'), verbose = self.verbose, trap_signals = None)
 
 				# clear the segment lists in place
@@ -408,6 +436,13 @@ class Handler(simplehandler.Handler):
 				print >>sys.stderr, "Warning: couldn't build segment list on checkpoint, probably there aren't any segments"
 
 	def _gatehandler(self, elem, timestamp, (segtype, instrument, new_state)):
+		# FIXME:  this method could conceivably be patched to know
+		# what to do with state transitions that appear to have
+		# gone backwards, i.e., it could correct segments that are
+		# already in the segmentlist.  this might be one way to
+		# sort out the problem of segment state snap-shotting code
+		# artificially claiming segments to be on beyond the time
+		# when they should stop.
 		timestamp = LIGOTimeGPS(0, timestamp)	# timestamp is in nanoseconds
 		state_key = (segtype, instrument)
 
@@ -462,6 +497,59 @@ class Handler(simplehandler.Handler):
 			output.close()
 		return outstr
 
+	def update_cumulative_segments(self):
+		"""!
+		A method to update the cumulative segment list 
+		"""
+		# FIXME Type casts should be removed when we switch to swig bindings
+		current_gps_time = float(lal.GPSTimeNow())
+		seglist_to_drop = segments.segmentlist([segments.segment(segments.NegInfinity, LIGOTimeGPS(current_gps_time - self.segment_history_duration))])
+		for segtype, seglistdict in self.cumulative_seglistdicts.items():
+			seglistdict.extend(self.seglistdicts[segtype])
+			seglistdict.coalesce()
+			for seglist in seglistdict.values():
+				seglist -= seglist_to_drop
+
+	def gen_cumulative_segments_xmldoc(self):
+		"""!
+		A method to output the cumulative segment list in a valid
+		ligolw xml format.
+		"""
+		xmldoc = ligolw.Document()
+		xmldoc.appendChild(ligolw.LIGO_LW())
+		process = ligolw_process.register_to_xmldoc(xmldoc, "gstlal_inspiral", {})
+
+		# Toggle segments off and on to make sure segment information
+		# added to the cumulative segments is current This needs to be
+		# run with self.dataclass.lock, but this function is only
+		# called currently by web_get_cumulative_segments_xml, which
+		# calls with with self.dataclass.lock
+		try:
+			# FIXME Timestamp here needs to be thought about more,
+			# for the same reason mentioned _gatehandler
+			timestamp = self.seglistdicts["triggersegments"].extent_all()[1].ns()
+			self._close_segments(timestamp)
+		except ValueError:
+			# no segments
+			print >>sys.stderr, "cannot close segments before updating cumulative segments, segment info may be incomplete"
+		self.update_cumulative_segments()
+		with ligolw_segments.LigolwSegments(xmldoc, process) as ligolwsegments:
+			for segtype, seglistdict in self.cumulative_seglistdicts.items():
+				ligolwsegments.insert_from_segmentlistdict(seglistdict, name = segtype, comment = "LLOID snapshot")
+		ligolw_process.set_process_end_time(process)
+		return xmldoc
+
+	def web_get_cumulative_segments_xml(self):
+		"""!
+		provide a bottle route to get cumulative segment information via a url
+		"""
+		with self.dataclass.lock:
+			output = StringIO.StringIO()
+			ligolw_utils.write_fileobj(self.gen_cumulative_segments_xmldoc(), output, trap_signals = None)
+			outstr = output.getvalue()
+			output.close()
+		return outstr
+
 	def gen_psd_xmldoc(self):
 		xmldoc = lalseries.make_psd_xmldoc(self.psds)
 		process = ligolw_process.register_to_xmldoc(xmldoc, "gstlal_inspiral", {})
@@ -565,20 +653,15 @@ def mkLLOIDbranch(pipeline, src, bank, bank_fragment, (control_snk, control_src)
 	# convolution, high-frequency branches use FFT convolution with a
 	# block stride given by fir_stride.
 	#
-	# FIXME:  why the -1?  without it the pieces don't match but I
-	# don't understand where this offset comes from.  it might really
-	# need to be here, or it might be a symptom of a bug elsewhere.
-	# figure this out.
 
-	latency = -int(round(bank_fragment.start * bank_fragment.rate)) - 1
+	latency = -int(round(bank_fragment.start * bank_fragment.rate))
 	block_stride = fir_stride * bank_fragment.rate
-	
+
 	# we figure an fft costs ~5 logN flops where N is duration + block
-	# stride.  For each chunk you have to do a forward and a reverse fft.
-	# Time domain costs N * block_stride. So if block stride is less than
-	# about 10logN you might as well do time domain filtering
+	# stride.  Time domain costs N * block_stride. So if block stride is
+	# less than about 5logN you might as well do time domain filtering
 	# FIXME This calculation should probably be made more rigorous
-	time_domain = 10 * numpy.log2((bank_fragment.end - bank_fragment.start) * bank_fragment.rate + block_stride) > block_stride
+	time_domain = 5 * numpy.log2((bank_fragment.end - bank_fragment.start) * bank_fragment.rate + block_stride) > block_stride
 
 	src = pipeparts.mkfirbank(pipeline, src, latency = latency, fir_matrix = bank_fragment.orthogonal_template_bank, block_stride = block_stride, time_domain = time_domain)
 	src = pipeparts.mkchecktimestamps(pipeline, src, "timestamps_%s_after_firbank" % logname)
@@ -615,17 +698,22 @@ def mkLLOIDbranch(pipeline, src, bank, bank_fragment, (control_snk, control_src)
 		# signal if that element gets smarter maybe this could be made smaller
 		# It should be > 1 * control_peak_time * gst.SECOND + 4 * block_duration
 		#
+		# FIXME for an unknown reason there needs to be an extra large
+		# queue in this part of the pipeline in order to prevent
+		# lock-ups.  Fortunately this is buffering up relatively
+		# lightweight data, i.e., before reconstruction
+		#
 		# FIXME since peakfinding is done, or control is based on
 		# injections only, we ignore the bank.gate_threshold parameter
 		# and just use 1e-100
 
 		src = pipeparts.mkgate(
 			pipeline,
-			pipeparts.mkqueue(pipeline, src, max_size_buffers = 0, max_size_bytes = 0, max_size_time = 1 * (2 * control_peak_time + (abs(gate_attack_length) + abs(gate_hold_length)) / bank_fragment.rate) * gst.SECOND + 10 * block_duration),
+			pipeparts.mkqueue(pipeline, src, max_size_buffers = 0, max_size_bytes = 0, max_size_time = 1 * (2 * control_peak_time + (abs(gate_attack_length) + abs(gate_hold_length)) / bank_fragment.rate) * gst.SECOND + 12 * block_duration),
 			threshold = 1e-100,
 			attack_length = gate_attack_length,
 			hold_length = gate_hold_length,
-			control = pipeparts.mkqueue(pipeline, control_src, max_size_buffers = 0, max_size_bytes = 0, max_size_time = 1 * (2 * control_peak_time + (abs(gate_attack_length) + abs(gate_hold_length)) / bank_fragment.rate) * gst.SECOND + 10 * block_duration)
+			control = pipeparts.mkqueue(pipeline, control_src, max_size_buffers = 0, max_size_bytes = 0, max_size_time = 1 * (2 * control_peak_time + (abs(gate_attack_length) + abs(gate_hold_length)) / bank_fragment.rate) * gst.SECOND + 12 * block_duration)
 		)
 		src = pipeparts.mkchecktimestamps(pipeline, src, "timestamps_%s_after_gate" % logname)
 	else:
@@ -972,6 +1060,9 @@ def mkLLOIDmulti(pipeline, detectors, banks, psd, psd_fft_length = 8, ht_gate_th
 			snrslices = snrslices
 		)
 		snr = pipeparts.mkchecktimestamps(pipeline, snr, "timestamps_%s_snr" % suffix)
+		# uncomment this tee if the diagnostic sinks below are
+		# needed
+		#snr = pipeparts.mktee(pipeline, snr)
 		if chisq_type == 'autochisq':
 			# FIXME don't hardcode
 			# peak finding window (n) in samples is 1 second at max rate, ie max(rates)
